package exam8.solution;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSortedSet;

import java.util.*;

interface Expression {
    double getValue(EvaluationContext context);

    Expression copy();

    Expression derivative(String variable);

    void accept(ExpressionVisitor visitor);

    Iterator<String> variableIterator();
}

class EvaluationContext {
    final private Map<String, Double> memory = new HashMap<>();

    void setVariable(String name, double val) {
        memory.put(name, val);
    }

    double getVariable(String name) {
        if (memory.containsKey(name)) {
            return memory.get(name);
        } else {
            throw new IllegalStateException("Variable \"" + name + "\" not set !");
        }
    }

}

abstract class AbstractExpression implements Expression {
    @Override
    public Iterator<String> variableIterator() {
        return new VariableIterator(this);
    }

    static Expression multiplyByInnerDerivation(String variable, Expression outer, Expression inner) {
        return new Multiply(outer, inner.derivative(variable));
    }

    @Override
    public String toString() {
        PrintExpressionVisitor v = new PrintExpressionVisitor();
        accept(v);
        return v.getResult();
    }
}

abstract class CompoundExpression extends AbstractExpression {
    final protected ImmutableList<Expression> args;

    protected CompoundExpression(ImmutableList<Expression> args) {
        this.args = args;
    }

    Expression getArg(int index) {
        return args.get(index);
    }

    abstract protected CompoundExpression create(ImmutableList<Expression> args);
}

class Constant extends AbstractExpression {
    final private double value;

    Constant(double value) {
        this.value = value;
    }

    @Override
    public double getValue(EvaluationContext context) {
        return value;
    }

    @Override
    public Expression derivative(String variable) {
        return new Constant(0.0);
    }

    @Override
    public void accept(ExpressionVisitor visitor) {
        visitor.visitConstant(this);
    }

    @Override
    public Expression copy() {
        return new Constant(value);
    }

    public Set<String> collectVariables() {
        return new HashSet<>();
    }
}

class Variable extends AbstractExpression {
    final private String name;

    Variable(String name) {
        this.name = name;
    }

    String getName() {
        return name;
    }

    @Override
    public double getValue(EvaluationContext context) {
        return context.getVariable(name);
    }

    @Override
    public Expression derivative(String variable) {
        return variable.equals(name) ? new Constant(1.0) : new Constant(0.0);
    }

    @Override
    public void accept(ExpressionVisitor visitor) {
        visitor.visitVariable(this);
    }

    @Override
    public Expression copy() {
        return new Variable(getName());
    }
}

class Sin extends CompoundExpression {
    Sin(Expression arg) {
        super(ImmutableList.<Expression>of(arg));
    }

    @Override
    protected CompoundExpression create(ImmutableList<Expression> args) {
        return new Sin(args.get(0));
    }

    @Override
    public double getValue(EvaluationContext context) {
        return Math.sin(getArg(0).getValue(context));
    }

    @Override
    public Expression derivative(String variable) {
        Expression inner = getArg(0).copy();
        return multiplyByInnerDerivation(variable, new Cos(inner), inner);
    }

    @Override
    public void accept(ExpressionVisitor visitor) {
        visitor.visitSin(this);
    }

    @Override
    public Expression copy() {
        return new Sin(getArg(0).copy());
    }
}

class Cos extends CompoundExpression {
    Cos(Expression arg) {
        super(ImmutableList.<Expression>of(arg));
    }

    @Override
    protected CompoundExpression create(ImmutableList<Expression> args) {
        return new Cos(args.get(0));
    }

    @Override
    public double getValue(EvaluationContext context) {
        return Math.cos(getArg(0).getValue(context));
    }

    @Override
    public Expression derivative(String variable) {
        Expression inner = getArg(0).copy();
        Expression outer = new Multiply(new Constant(-1.0), new Sin(inner));
        return multiplyByInnerDerivation(variable, outer, inner);
    }

    @Override
    public void accept(ExpressionVisitor visitor) {
        visitor.visitCos(this);
    }

    @Override
    public Expression copy() {
        return new Cos(getArg(0).copy());
    }
}

class Power extends CompoundExpression {

    Power(Expression base, Expression exp) {
        super(ImmutableList.<Expression>of(base, exp));
    }

    @Override
    protected CompoundExpression create(ImmutableList<Expression> args) {
        return new Power(args.get(0), args.get(1));
    }

    @Override
    public double getValue(EvaluationContext context) {
        return Math.pow(getArg(0).getValue(context), getArg(1).getValue(context));
    }

    @Override
    public Expression derivative(String variable) {
        Expression inner = getArg(0).copy();
        Expression exponent = getArg(1).copy();
        Expression exponentMinusOne = new Add(exponent, new Constant(-1.0));
        Expression outer = new Multiply(exponent, new Power(inner, exponentMinusOne));
        return multiplyByInnerDerivation(variable, outer, inner);
    }

    @Override
    public void accept(ExpressionVisitor visitor) {
        visitor.visitPower(this);
    }

    @Override
    public Expression copy() {
        return new Power(getArg(0).copy(), getArg(1).copy());
    }
}

class Add extends CompoundExpression {
    Add(ImmutableList<Expression> args) {
        super(args);
    }

    Add(Expression left, Expression right) {
        this(ImmutableList.<Expression>of(left, right));
    }

    @Override
    protected CompoundExpression create(ImmutableList<Expression> args) {
        return new Add(args);
    }

    @Override
    public double getValue(EvaluationContext context) {
        double sum = 0.0;
        for (Expression e : args) {
            sum += e.getValue(context);
        }
        return sum;
    }

    @Override
    public Expression derivative(String variable) {
        ImmutableList.Builder<Expression> b = ImmutableList.builder();
        for (Expression e : args) {
            b.add(e.derivative(variable));
        }
        return new Add(b.build());
    }

    @Override
    public void accept(ExpressionVisitor visitor) {
        visitor.visitAdd(this);
    }

    public Expression copy() {
        ImmutableList.Builder<Expression> b = ImmutableList.builder();
        for (Expression e : args) {
            b.add(e.copy());
        }
        return new Add(b.build());
    }
}

class Multiply extends CompoundExpression {

    Multiply(ImmutableList<Expression> args) {
        super(args);
    }

    Multiply(Expression left, Expression right) {
        this(ImmutableList.<Expression>of(left, right));
    }

    @Override
    protected CompoundExpression create(ImmutableList<Expression> args) {
        return new Multiply(args);
    }

    @Override
    public double getValue(EvaluationContext context) {
        double total = 1.0;
        for (Expression e : args) {
            total *= e.getValue(context);
        }
        return total;
    }

    @Override
    public Expression derivative(String variable) {
        return derivativeHelper(0, variable);
    }

    private Expression derivativeHelper(int i, String variable) {
        if (i < args.size() - 1) {
            Expression u = getArg(i).copy();
            Expression du = u.derivative(variable);
            ImmutableList.Builder<Expression> builder = ImmutableList.builder();
            for (int j = i + 1; j < args.size(); j++) {
                builder.add(getArg(j).copy());
            }
            ImmutableList<Expression> v = builder.build();
            Expression dv = derivativeHelper(i + 1, variable);
            return new Add(
                    new Multiply(ImmutableList.<Expression>builder().add(du).addAll(v).build()),
                    new Multiply(u, dv)
            );
        } else {
            return getArg(i).derivative(variable);
        }
    }


    @Override
    public void accept(ExpressionVisitor visitor) {
        visitor.visitMultiply(this);
    }

    public Expression copy() {
        ImmutableList.Builder<Expression> b = ImmutableList.builder();
        for (Expression e : args) {
            b.add(e.copy());
        }
        return new Multiply(b.build());
    }
}

interface ExpressionVisitor {
    void visitConstant(Constant c);

    void visitVariable(Variable v);

    void visitSin(Sin v);

    void visitCos(Cos v);

    void visitPower(Power v);

    void visitAdd(Add v);

    void visitMultiply(Multiply v);
}

class PrintExpressionVisitor implements ExpressionVisitor {
    private String result = null;

    @Override
    public void visitConstant(Constant c) {
        result = "" + c.getValue(null);
    }

    @Override
    public void visitVariable(Variable v) {
        result = v.getName();
    }

    @Override
    public void visitSin(Sin v) {
        v.getArg(0).accept(this);
        result = "sin(" + result + ")";
    }

    @Override
    public void visitCos(Cos v) {
        v.getArg(0).accept(this);
        result = "cos(" + result + ")";
    }

    @Override
    public void visitPower(Power v) {
        v.getArg(0).accept(this);
        String left = result;
        v.getArg(1).accept(this);
        String right = result;
        result = left + "^" + right;
    }

    @Override
    public void visitAdd(Add v) {
        //TODO StringBuilder
        v.getArg(0).accept(this);
        String res = "(" + result;
        for (int i = 1; i < v.args.size(); i++) {
            v.getArg(i).accept(this);
            res += " + " + result;
        }
        result = res + ")";
    }

    @Override
    public void visitMultiply(Multiply v) {
        //TODO StringBuilder
        v.getArg(0).accept(this);
        String res = "(" + result;
        for (int i = 1; i < v.args.size(); i++) {
            v.getArg(i).accept(this);
            res += " * " + result;
        }
        result = res + ")";
    }

    String getResult() {
        return result;
    }
}

class SimplifyExpressionVisitor implements ExpressionVisitor {
    private Expression result = null;

    Expression getResult() {
        return result;
    }

    private Expression simplifyAllConstants(CompoundExpression ce) {
        for (Expression e : ce.args) {
            if (!(e instanceof Constant)) {
                return ce;
            }
        }
        return new Constant(ce.getValue(null));
    }

    private CompoundExpression simplifySubExpressions(CompoundExpression ce) {
        ImmutableList.Builder<Expression> builder = ImmutableList.builder();
        for (Expression e : ce.args) {
            e.accept(this);
            builder.add(result);
        }
        return ce.create(builder.build());
    }

    private CompoundExpression gatherConstants(CompoundExpression ce) {
        ImmutableList.Builder<Expression> nonConsts = ImmutableList.builder();
        ImmutableList.Builder<Expression> consts = ImmutableList.builder();
        for (Expression e : ce.args) {
            if (e instanceof Constant) {
                consts.add(e);
            } else {
                nonConsts.add(e);
            }
        }
        Constant c = new Constant(ce.create(consts.build()).getValue(null));
        ImmutableList.Builder<Expression> out = ImmutableList.builder();
        return ce.create(out.addAll(nonConsts.build()).add(c).build());
    }

    private CompoundExpression removeConstant(CompoundExpression ce, double valueToRemove) {
        ImmutableList.Builder<Expression> builder = ImmutableList.builder();
        for (Expression e : ce.args) {
            if (!(e instanceof Constant && e.getValue(null) == valueToRemove)) {
                builder.add(e);
            }
        }
        return ce.create(builder.build());
    }

    @Override
    public void visitConstant(Constant c) {
        result = c.copy();
    }

    @Override
    public void visitVariable(Variable v) {
        result = v.copy();
    }

    @Override
    public void visitSin(Sin v) {
        result = simplifyAllConstants(simplifySubExpressions(v));
    }

    @Override
    public void visitCos(Cos v) {
        result = simplifyAllConstants(simplifySubExpressions(v));
    }

    @Override
    public void visitPower(Power v) {
        CompoundExpression ce = simplifySubExpressions(v);
        if (ce.getArg(0) instanceof Constant) {
            double base = ce.getArg(0).getValue(null);
            if (base == 0.0) {
                result = new Constant(0.0);
                return;
            } else if (base == 1.0) {
                result = new Constant(1.0);
                return;
            }
        }

        if (ce.getArg(1) instanceof Constant) {
            double exp = ce.getArg(1).getValue(null);
            if (exp == 0.0) {
                result = new Constant(1.0);
                return;
            } else if (exp == 1.0) {
                result = ce.getArg(0);
                return;
            }
        }
        result = simplifyAllConstants(ce);
    }

    @Override
    public void visitAdd(Add v) {
        CompoundExpression ce = simplifySubExpressions(v);
        CompoundExpression gc = gatherConstants(ce);
        CompoundExpression rc = removeConstant(gc, 0.0);
        result = simplifyAllConstants(rc);
    }

    @Override
    public void visitMultiply(Multiply v) {
        CompoundExpression ce = simplifySubExpressions(v);
        for (Expression e : ce.args) {
            if (e instanceof Constant && e.getValue(null) == 0.0) {
                result = new Constant(0.0);
                return;
            }
        }
        CompoundExpression gc = gatherConstants(ce);
        CompoundExpression rc = removeConstant(gc, 1.0);
        result = simplifyAllConstants(rc);
    }
}

class VariableCollectVisitor implements ExpressionVisitor {
    private Set<String> result = null;

    Set<String> getResult() {
        return result;
    }

    @Override
    public void visitConstant(Constant c) {
        result = new HashSet<>();
    }

    @Override
    public void visitVariable(Variable v) {
        result = new HashSet<>();
        result.add(v.getName());
    }

    private void collectCompound(CompoundExpression c) {
        Set<String> variableSet = new HashSet<>();
        for (Expression e : c.args) {
            e.accept(this);
            variableSet.addAll(result);
        }
        result = variableSet;
    }

    @Override
    public void visitSin(Sin v) {
        collectCompound(v);
    }

    @Override
    public void visitCos(Cos v) {
        collectCompound(v);
    }

    @Override
    public void visitPower(Power v) {
        collectCompound(v);
    }

    @Override
    public void visitAdd(Add v) {
        collectCompound(v);
    }

    @Override
    public void visitMultiply(Multiply v) {
        collectCompound(v);
    }
}

class VariableIterator implements Iterator<String> {
    final private Iterator<String> iterator;

    VariableIterator(Expression e) {
        VariableCollectVisitor v = new VariableCollectVisitor();
        e.accept(v);
        this.iterator = ImmutableSortedSet.copyOf(v.getResult()).iterator();

    }

    @Override
    public boolean hasNext() {
        return iterator.hasNext();
    }

    @Override
    public String next() {
        return iterator.next();
    }

    @Override
    public void remove() {
        //throw new IllegalStateException("Operation not supported: remove()");
        iterator.remove();
    }
}

public class codeSolution {
    public static void main(String[] args) {
        SimplifyExpressionVisitor simplifyExpressionVisitor = new SimplifyExpressionVisitor();

        EvaluationContext context = new EvaluationContext();
        context.setVariable("x", 3.0);
        context.setVariable("y", -3.0);

//        Expression e = new Variable("x");
//        Expression e = new Sin(new Add(ImmutableList.<Expression>of(new Variable("x"), new Constant(0.0), new Constant(2.0), new Constant(0.0))));
//        Expression e = new Add(ImmutableList.<Expression>of(new Variable("x"), new Constant(0.0), new Constant(2.0)));
//        Expression e = new Multiply(ImmutableList.<Expression>of(new Variable("x"), new Constant(2), new Constant(3)));
//        Expression e = new Power(new Variable("x"), new Add(new Constant(2.0), new Constant(-1.0)));
        Expression e = new Power(new Add(new Constant(2.0), new Constant(-2.0)), new Variable("x"));

//        Expression e = new Multiply(new Constant(5.0), new Sin(new Power(new Variable("x"), new Constant(2))));
//        Expression e = new Multiply(ImmutableList.<Expression>of(
//                new Variable("x"),
//                new Sin(new Variable("x")),
//                new Cos(new Variable("x"))
//        ));
        System.out.println("original:              " + e);

        e.accept(simplifyExpressionVisitor);
        Expression se = simplifyExpressionVisitor.getResult();
        System.out.println("simplified  original:  " + se);


//        for (Iterator<String> i = e1.variableIterator(); i.hasNext(); ) {
//            String s = i.next();
//            System.out.println(s);
//        }


        Expression d = e.derivative("x");
        System.out.println("derivative:            " + d);

        d.accept(simplifyExpressionVisitor);
        Expression sd = simplifyExpressionVisitor.getResult();
        System.out.println("simplified derivative: " + sd);

        System.out.println(sd.getValue(context));
    }
}
